#include <math.h>
#include <iostream>
using namespace std;

double* fwt_decomp(double signal[],int siglen) {
	bool is_2power = false;
	double a[] = {1.000, 1.000};
	double b[] = {1.000, -1.000};
	int filter_length = sizeof(a)/sizeof(double);
	/*for(int p=0;pow(2,p)<=siglen;p++) {
		if(siglen == pow(2,p))  {
			is_2power = true;
		}
	}
	if(!is_2power) {
		return NULL;
	}*/
	int c_size = siglen;
	int decomp_length = 0;
	int iter_length = siglen;
	int levels = 1;
	while(iter_length > filter_length) {
		iter_length = ceil((double)iter_length/2.) + ceil((double)filter_length/2. - 1.);
		decomp_length += iter_length;
		levels++;
	}
	decomp_length += iter_length;
	int *d_index = new int[levels];
	d_index[levels-1] = levels+2;
	iter_length = siglen;
	int offset = decomp_length+levels+2;
	levels = 1;
	while(iter_length > filter_length) {
		if(levels != 1) {
			offset = d_index[levels-2];
		}
		iter_length = ceil((double)iter_length/2.) + ceil((double)filter_length/2. - 1.);
		d_index[levels-1] = offset-iter_length;
		levels++;
	}
	double *decomp = new double[decomp_length+levels+2];
	decomp[0] = levels;
	decomp[1] = decomp_length;
	for(int i=1;i<=levels;i++) {
		decomp[i+1] = d_index[i-1];
	}
	double *c,*d;
	c = new double[siglen];
	for(int i=0;i<siglen;i++) {
		c[i] = signal[i];
	}
	levels = 1;
	while(true)
	{
		int dim = ceil((double)c_size/2.) + ceil(filter_length/2 - 1);
		double *new_c = new double[dim];
		d = new double[dim];
		new_c[dim-1] = 0;
		d[dim-1] = 0;

		// Haar - D2:
		// Low-pass:
		/*for(int k=1;2*k-1 < c_size;k++) {
			new_c[k-1] = sqrt(2.)*(c[2*k-1]+c[2*k-2])/2.;
		}
		// High-pass:
		for(int k=1;2*k-1 < c_size;k++) {
			d[k-1] = sqrt(2.)*(c[2*k-2]-c[2*k-1])/2.;
			decomp[c_size/2+k-1] = d[k-1];
		}*/

		// D4:
		for(int i=1; i <= dim;i++) {
			new_c[i-1] = 0.;
			d[i-1] = 0.;
			for(int k=1;k <= filter_length;k++) {
				if(2*i-(filter_length-k) >= 1 && 2*i-(filter_length-k) <= c_size) {
					// Low-pass:
					new_c[i-1] += a[filter_length - k]*c[2*i-(filter_length - k)-1];
					// High-pass:
					d[i-1] += b[filter_length - k]*c[2*i-(filter_length - k)-1];
				}
			}
			int index = d_index[levels-1] + i - 1;
			decomp[index] = d[i-1];
		}
		delete[] c;
		delete[] d;
		c = new_c;
		levels++;
		c_size = dim;
		if(dim <= filter_length) {
			break;
		}
		
	}
	for(int i=1;i <= c_size;i++) {
		int index = d_index[levels-1] + i - 1;
		decomp[index] = c[i-1];
	}
	delete[] c;
	delete[] d_index;
	return decomp;
}

double* fwt_reconst(double decomp[],int ret_len) {
	int siglen = decomp[1];
	if(ret_len > siglen) {
		return NULL;
	}
	bool is_2power = false;
	int levels = decomp[0];
	int *d_index = new int[levels+1];
	for(int i=0;i<levels;i++) {
		d_index[i] = decomp[levels+1-i]-levels-2;
	}
	d_index[levels] = siglen;
	double a[] = {1.000, 1.000};
	double b[] = {1.000, -1.000};
	//double a[] = {0.6830127, 1.1830127, 0.3169873, -0.1830127};
	//double b[] = {-0.1830127, -0.3169873, 1.1830127, -0.6830127};
	double *copy = new double[siglen];
	int filter_length = sizeof(a)/sizeof(double);
	/*for(int p=0;pow(2,p)<=siglen;p++) {
		if(siglen == pow(2,p))  {
			is_2power = true;
		}
	}
	if(!is_2power) {
		return NULL;
	}*/
	for(int i=0;i<siglen;i++) {
		copy[i] = decomp[i+levels+2];
	}
	int start_level = levels;
	while(levels > 1)
	{
		int level_length_low = d_index[start_level - levels + 1];
		int level_length_high = d_index[start_level-levels+2]-d_index[start_level - levels +1];//cout << start_level-levels+2 << endl;
		int level_length = level_length_high;
		int new_c_len = 2*level_length;//cout << "-2 " << level_length << endl;
		double *new_c = new double[new_c_len];
		for(int l=1;l<=2*level_length;l++) {
			new_c[l-1] = 0;
			for(int k=1;k<=level_length;k++) {
				if(level_length_low >= k && 2*k-(l-1) >= 1 && 2*k-(l-1) <= filter_length) {
					new_c[l-1] += a[2*k-l]*copy[k-1];
				}
				if(k >= 1 && level_length_high >= k && 2*k-(l-1) >= 1 && 2*k-(l-1) <= filter_length) {
					new_c[l-1] += b[2*k-l]*copy[k + d_index[start_level-levels+1]-1];
				}
			}
		}
		for(int i=0;i < new_c_len;i++) {
			if(i < level_length_low + level_length_high) {
				copy[i] = new_c[i]/2.;
			}
		}
		delete[] new_c;
		levels--;
	}
	double *reconst_signal = new double[ret_len];
	for(int i=0;i<ret_len;i++) {
		reconst_signal[i] = copy[i];
	}
	/*for(int j=1;(int)pow(2,j) <= siglen;j++)
	{
		double *new_c = new double[(int)pow(2,j)];
		int d_offset = pow(2,j-1);

		// Haar - D2:
		for(int k=1;2*k <= pow(2,j);k++) {
			new_c[2*k-1] = sqrt(2.)*(decomp[k-1]-decomp[d_offset+(k-1)])/2.;
			new_c[2*k-2] = sqrt(2.)*(decomp[k-1]+decomp[d_offset+(k-1)])/2.;
		}
		for(int k=1;k <= pow(2,j);k++) {
			decomp[k-1] = new_c[k-1];
		}

		// D4:
		for(int k=1;2*k <= pow(2,j);k++) {
			new_c[2*k-1] = 0;
			new_c[2*k-2] = 0;
		}
		for(int k=1;k <= pow(2,j);k++) {
			decomp[k-1] = new_c[k-1];
		}
		delete[] new_c;
	}*/
	delete[] copy;
	delete[] d_index;
	return reconst_signal;
}
