#include <iostream>
#include <sstream>
#include <math.h>
#include "EasyBMP.h"
#include "dwt.h"
#include "raster_sort.h"

using namespace std;

int main(int argc, char **argv)
{
	if(argc != 2) {
		cout << "Usage: wavelet_image <filename>" << endl;
		return 1;
	}
	BMP input;
	input.ReadFromFile(argv[1]);
	const int width = input.TellWidth();
	const int height = input.TellHeight();
	int fw = 16;
	int fh = 16;
	
	
	if(width%fw != 0 || height%fh != 0) {
		cout << "Image-width and -height must be a multiple of 8" << endl;
		return 1;
	}
	cout.setf(ios::fixed,ios::floatfield);
	cout.precision(4);

	//
	double *test_signal = new double[fw*fh];
	for(int i=0;i< fw*fh;i++) {
		test_signal[i] = i;
	}
	double *ret_signal = fwt_decomp(test_signal,fw*fh);
	delete[] test_signal;
	int levels = ret_signal[0];
	int decomp_length = levels+ret_signal[1]+2;
	int *d_index = new int[levels+1];
	for(int i=0;i<=levels;i++) {
		d_index[i] = ret_signal[i+1];
	}
	int fields = width*height/(fw*fh);
	double *transform_array = new double[fields*decomp_length];
	for(int j=0;j<height/fh;j++) {
		for(int i=0;i<width/fw;i++) {
			double *signal = new double[fw*fh];
			for(int k=0;k<fw*fh;k++) {
				int l = i*fw+k%fw;
				int m = j*fh+(k-(k%fw))/fw;
				int color = (int)input(l,m)->Red;
				signal[k] = color;
			}
			//raster_sort(signal,fw*fh,fw,fh);
			double *decomp_signal;
			decomp_signal = fwt_decomp(signal,fw*fh);
			/*for(int c1=0;c1<=fw*fh;c1++)
			{
				cout << signal[c1] << "\t";
			}
			cout << endl << endl;
			for(int c1=0;c1<=decomp_length;c1++)
			{
				cout << decomp_signal[c1] << "\t";
			}
			cout << endl << endl;
			break;*/
			for(int n=0;n<decomp_length;n++) {
				int offset = i+j*width/fw;
				transform_array[n*fields+offset] = decomp_signal[n+levels+2];
			}
			delete[] decomp_signal;
			delete[] signal;
		}
		//break;
	}
	
	// Reconstructed images:
	for(int p=1;p<=levels;p++) {
		int approx = d_index[levels-p];
		BMP output;
		output.SetSize(width,height);
		for(int j=0;j<height/fh;j++) {
			for(int i=0;i<width/fw;i++) {
				double *input_signal = new double[decomp_length];
				input_signal[0] = levels;
				for(int c1=1;c1<=levels+1;c1++) {
					input_signal[c1] = d_index[c1-1];
				}
				for(int n=0;n<approx;n++) {
					int offset = i+j*width/fw;
					input_signal[n+levels+2] = transform_array[n*fields+offset];
				}
				for(int n=approx;n<decomp_length;n++) {
					input_signal[n] = 0.;
				}
				double *reconst_signal;
				reconst_signal = fwt_reconst(input_signal,fw*fh);
				//inv_raster_sort(reconst_signal,fw*fh,fw,fh);
				for(int n=0;n<fw*fh;n++) {
					int l = i*fw+n%fw;
					int m = j*fh+(n-(n%fw))/fw;
					output(l,m)->Red = reconst_signal[n];
					output(l,m)->Green = reconst_signal[n];
					output(l,m)->Blue = reconst_signal[n];
				}
				delete[] input_signal;
				delete[] reconst_signal;
			}
		}
		char filename[54];
		stringstream ss;
		ss << "approx_" <<  approx << ".bmp";
		ss >> filename;
		output.WriteToFile(filename);
		cout << filename << " created!" << endl;
	}
	delete[] d_index;
	delete[] transform_array;
	
	return 0;
}
